@@ -19,14 +19,14 @@ import datetime
 import inspect
 import os
 import warnings
+from collections.abc import Collection, Generator, Sequence
 from enum import Enum, EnumMeta, IntEnum, unique
 from typing import (
     TYPE_CHECKING,
     Any,
+    Callable,
     ClassVar,
-    Collection,
-    Dict,
-    List,
+    NoReturn,
     Optional,
     TypeVar,
     Union,
@@ -45,17 +45,20 @@ from hypothesis.utils.conventions import not_set
 from hypothesis.utils.dynamicvariables import DynamicVariable
 
 if TYPE_CHECKING:
+    from typing import TypeAlias
+
     from hypothesis.database import ExampleDatabase
 
 __all__ = ["settings"]
 
-all_settings: Dict[str, "Setting"] = {}
+ValidatorT: "TypeAlias" = Callable[[Any], object]
+all_settings: dict[str, "Setting"] = {}
 
 T = TypeVar("T")
 
 
 class settingsProperty:
-    def __init__(self, name, show_default):
+    def __init__(self, name: str, *, show_default: bool) -> None:
         self.name = name
         self.show_default = show_default
 
@@ -72,6 +75,7 @@ class settingsProperty:
                     from hypothesis.database import ExampleDatabase
 
                     result = ExampleDatabase(not_set)
+                assert result is not not_set
                 return result
             except KeyError:
                 raise AttributeError(self.name) from None
@@ -93,7 +97,7 @@ class settingsProperty:
         return f"{description}\n\ndefault value: ``{default}``"
 
 
-default_variable = DynamicVariable(None)
+default_variable = DynamicVariable[Optional["settings"]](None)
 
 
 class settingsMeta(type):
@@ -101,19 +105,20 @@ class settingsMeta(type):
         super().__init__(*args, **kwargs)
 
     @property
-    def default(cls):
+    def default(cls) -> Optional["settings"]:
         v = default_variable.value
         if v is not None:
             return v
-        if hasattr(settings, "_current_profile"):
+        if getattr(settings, "_current_profile", None) is not None:
+            assert settings._current_profile is not None
             settings.load_profile(settings._current_profile)
             assert default_variable.value is not None
         return default_variable.value
 
-    def _assign_default_internal(cls, value):
+    def _assign_default_internal(cls, value: "settings") -> None:
         default_variable.value = value
 
-    def __setattr__(cls, name, value):
+    def __setattr__(cls, name: str, value: object) -> None:
         if name == "default":
             raise AttributeError(
                 "Cannot assign to the property settings.default - "
@@ -126,7 +131,7 @@ class settingsMeta(type):
                 "settings with settings.load_profile, or use @settings(...) "
                 "to decorate your test instead."
             )
-        return super().__setattr__(name, value)
+        super().__setattr__(name, value)
 
 
 class settings(metaclass=settingsMeta):
@@ -138,8 +143,9 @@ class settings(metaclass=settingsMeta):
     """
 
     __definitions_are_locked = False
-    _profiles: ClassVar[Dict[str, "settings"]] = {}
+    _profiles: ClassVar[dict[str, "settings"]] = {}
     __module__ = "hypothesis"
+    _current_profile = None
 
     def __getattr__(self, name):
         if name in all_settings:
@@ -240,14 +246,14 @@ class settings(metaclass=settingsMeta):
     @classmethod
     def _define_setting(
         cls,
-        name,
-        description,
+        name: str,
+        description: str,
         *,
-        default,
-        options=None,
-        validator=None,
-        show_default=True,
-    ):
+        default: object,
+        options: Optional[Sequence[object]] = None,
+        validator: Optional[ValidatorT] = None,
+        show_default: bool = True,
+    ) -> None:
         """Add a new setting.
 
         - name is the name of the property that will be used to access the
@@ -280,16 +286,16 @@ class settings(metaclass=settingsMeta):
             default=default,
             validator=validator,
         )
-        setattr(settings, name, settingsProperty(name, show_default))
+        setattr(settings, name, settingsProperty(name, show_default=show_default))
 
     @classmethod
-    def lock_further_definitions(cls):
+    def lock_further_definitions(cls) -> None:
         settings.__definitions_are_locked = True
 
-    def __setattr__(self, name, value):
+    def __setattr__(self, name: str, value: object) -> NoReturn:
         raise AttributeError("settings objects are immutable")
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         from hypothesis.internal.conjecture.data import AVAILABLE_PROVIDERS
 
         bits = sorted(
@@ -299,7 +305,7 @@ class settings(metaclass=settingsMeta):
         )
         return "settings({})".format(", ".join(bits))
 
-    def show_changed(self):
+    def show_changed(self) -> str:
         bits = []
         for name, setting in all_settings.items():
             value = getattr(self, name)
@@ -324,9 +330,15 @@ class settings(metaclass=settingsMeta):
         :class:`~hypothesis.settings`: optional ``parent`` settings, and
         keyword arguments for each setting that will be set differently to
         parent (or settings.default, if parent is None).
+
+        If you register a profile that has already been defined and that profile
+        is the currently loaded profile, the new changes will take effect immediately,
+        and do not require reloading the profile.
         """
         check_type(str, name, "name")
         settings._profiles[name] = settings(parent=parent, **kwargs)
+        if settings._current_profile == name:
+            settings.load_profile(name)
 
     @staticmethod
     def get_profile(name: str) -> "settings":
@@ -351,20 +363,20 @@ class settings(metaclass=settingsMeta):
 
 
 @contextlib.contextmanager
-def local_settings(s):
+def local_settings(s: settings) -> Generator[settings, None, None]:
     with default_variable.with_value(s):
         yield s
 
 
 @attr.s()
 class Setting:
-    name = attr.ib()
-    description = attr.ib()
-    default = attr.ib()
-    validator = attr.ib()
+    name: str = attr.ib()
+    description: str = attr.ib()
+    default: object = attr.ib()
+    validator: ValidatorT = attr.ib()
 
 
-def _max_examples_validator(x):
+def _max_examples_validator(x: int) -> int:
     check_type(int, x, name="max_examples")
     if x < 1:
         raise InvalidArgument(
@@ -394,7 +406,7 @@ running time against the chance of missing a bug.
 If you are writing one-off tests, running tens of thousands of examples is
 quite reasonable as Hypothesis may miss uncommon bugs with default settings.
 For very complex code, we have observed Hypothesis finding novel bugs after
-*several million* examples while testing :pypi:`SymPy`.
+*several million* examples while testing :pypi:`SymPy <sympy>`.
 If you are running more than 100k examples for a test, consider using our
 :ref:`integration for coverage-guided fuzzing <fuzz_one_input>` - it really
 shines when given minutes or hours to run.
@@ -416,11 +428,13 @@ This allows you to `check for regressions and look for bugs
 :ref:`separate settings profiles <settings_profiles>` - for example running
 quick deterministic tests on every commit, and a longer non-deterministic
 nightly testing run.
+
+By default when running on CI, this will be set to True.
 """,
 )
 
 
-def _validate_database(db):
+def _validate_database(db: "ExampleDatabase") -> "ExampleDatabase":
     from hypothesis.database import ExampleDatabase
 
     if db is None or isinstance(db, ExampleDatabase):
@@ -458,7 +472,7 @@ class Phase(IntEnum):
     shrink = 4  #: controls whether examples will be shrunk.
     explain = 5  #: controls whether Hypothesis attempts to explain test failures.
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"Phase.{self.name}"
 
 
@@ -475,14 +489,14 @@ class HealthCheck(Enum, metaclass=HealthCheckMeta):
     Each member of this enum is a type of health check to suppress.
     """
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"{self.__class__.__name__}.{self.name}"
 
     @classmethod
-    def all(cls) -> List["HealthCheck"]:
+    def all(cls) -> list["HealthCheck"]:
         # Skipping of deprecated attributes is handled in HealthCheckMeta.__iter__
         note_deprecation(
-            "`Healthcheck.all()` is deprecated; use `list(HealthCheck)` instead.",
+            "`HealthCheck.all()` is deprecated; use `list(HealthCheck)` instead.",
             since="2023-04-16",
             has_codemod=True,
             stacklevel=1,
@@ -556,7 +570,7 @@ class Verbosity(IntEnum):
     verbose = 2
     debug = 3
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"Verbosity.{self.name}"
 
 
@@ -568,7 +582,7 @@ settings._define_setting(
 )
 
 
-def _validate_phases(phases):
+def _validate_phases(phases: Sequence[Phase]) -> Sequence[Phase]:
     phases = tuple(phases)
     for a in phases:
         if not isinstance(a, Phase):
@@ -587,7 +601,7 @@ settings._define_setting(
 )
 
 
-def _validate_stateful_step_count(x):
+def _validate_stateful_step_count(x: int) -> int:
     check_type(int, x, name="stateful_step_count")
     if x < 1:
         raise InvalidArgument(f"stateful_step_count={x!r} must be at least one.")
@@ -645,12 +659,14 @@ settings._define_setting(
 class duration(datetime.timedelta):
     """A timedelta specifically measured in milliseconds."""
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         ms = self.total_seconds() * 1000
         return f"timedelta(milliseconds={int(ms) if ms == int(ms) else ms!r})"
 
 
-def _validate_deadline(x):
+def _validate_deadline(
+    x: Union[int, float, datetime.timedelta, None]
+) -> Optional[duration]:
     if x is None:
         return x
     invalid_deadline_error = InvalidArgument(
@@ -691,6 +707,8 @@ errors (but will not necessarily be if close to the deadline, to allow some
 variability in test run time).
 
 Set this to ``None`` to disable this behaviour entirely.
+
+By default when running on CI, this will be set to None.
 """,
 )
 
@@ -703,18 +721,16 @@ def is_in_ci() -> bool:
 
 settings._define_setting(
     "print_blob",
-    default=is_in_ci(),
-    show_default=False,
+    default=False,
     options=(True, False),
     description="""
 If set to ``True``, Hypothesis will print code for failing examples that can be used with
 :func:`@reproduce_failure <hypothesis.reproduce_failure>` to reproduce the failing example.
-The default is ``True`` if the ``CI`` or ``TF_BUILD`` env vars are set, ``False`` otherwise.
 """,
 )
 
 
-def _backend_validator(value):
+def _backend_validator(value: str) -> str:
     from hypothesis.internal.conjecture.data import AVAILABLE_PROVIDERS
 
     if value not in AVAILABLE_PROVIDERS:
@@ -759,6 +775,23 @@ def note_deprecation(
 
 settings.register_profile("default", settings())
 settings.load_profile("default")
+
+assert settings.default is not None
+
+CI = settings(
+    derandomize=True,
+    deadline=None,
+    database=None,
+    print_blob=True,
+    suppress_health_check=[HealthCheck.too_slow],
+)
+
+settings.register_profile("ci", CI)
+
+
+if is_in_ci():
+    settings.load_profile("ci")
+
 assert settings.default is not None
 
 
